#ifndef	_XPF_INTERNAL_H_
#define _XPF_INTERNAL_H_

#include <stdio.h>
#include "xpf_interface.h"

/*
 * Return values of comparison functions between currentpkt and given filter.
 */
#define XPF_DAGS_DESTPREFIX	-4	/* dest is a prefix of src */
#define XPF_DAGS_SRCPREFIX	-3	/* src is a prefix of dest */
#if	DELAYEDMSGS
#define XPF_DAGS_DELAY		-2	/* delay the message */
#endif	DELAYEDMSGS
/* Cannot change BEFORE, SAME and AFTER since utilities depend on it */
#define XPF_DAGS_BEFORE		-1	/* src must be matched before dest */
#define XPF_DAGS_SAME		0	/* src=dest */
#define XPF_DAGS_AFTER		1	/* src must be matched after dest */
#define XPF_DAGS_ERROR		2	/* header / extraction error */
#define XPF_DAGS_SAMEMASK	3	/* don't care if values =, we'll hash */

/*
 * We now define the structure which is a unification of the normal patricia
 * tree node containing and/or branches -- a node that can be split if another
 * node with a common prefix overlapps it, and a hash-table entry which has
 * just one and branch and a collision resolution next pointer which cannot
 * be split and has to match as a whole. This is distinguished by the nodetype.
 *
 * NOTE: a hash-entry can never directly point at another hash-entry: this
 * is used in all routines in xpf_dags.c in the sense that they can assume
 * that their parameters are always non-HASHENTRYs: though their return
 * values need not be. LaunchProcesses cannot assume this about matched nodes.
 */

/* This struct is not refered to anywhere except in PATNode */
typedef struct __hash {
	XPFDataUnit	value[XPF_DATA_MAXSIZE];
	XPFProgram	*programlist;
	short		alarm;
	short		processnum;
} _hash;

/* INVARIANT: a node can have EITHER ALL STATIC OR ALL DYNAMIC patterns */
#define XPF_FLAG_USER		0x3	/* & s.t. user can't goof w/ my flags */
#define XPF_FLAG_PROCESS	0x4	/* use in OrderPat(), set in Launch() */
#define XPF_DAGS_PROCESS	XPF_FLAG_PROCESS
#define XPF_DAGS_HASHENTRY	0x8	/* set in xpfInsertPatternList() */
/* There is no XPF_FLAG_HASHENTRY; _DAGS_ are used in nodes, _FLAG_ in pats */
#define XPF_FLAG_DAGS		0xf	/* My mask for all flags */
#define XPF_FLAG_RESIDUE	0xc	/* which remain with me if I split */

typedef struct _PATNode {
	int		refcount;
#if	DELAYEDMSGS && STOREMSGS
	XPFMessageList	*messagelist;	/* delayed messages */
#endif	DELAYEDMSGS && STOREMSGS
#if	MERGEINFO
	XPFInfoList	*info;		/* set of informations */
#else	MERGEINFO
	XPFInformation	info;		/* Unique information */
#endif	MERGEINFO
	struct _PATNode	*or;		/* Sequential or: Go here if it doesn't match */
	struct _PATNode	*and;		/* Sequential and: Go here if it does match */
	int		nodetype;	/* HASHENTRY or not (wasted, but what to do?) */

	union {
		XPFPattern	pattern;
		_hash		hash;
	} common;
} PATNode;

/*
 * These 3 attributes uniquely identify a position in the PAT DAG.
 * "Pointers to pointers" are used since "pointers" can be altered.
 * This is used to pass parameters in all "PAT DAG functions".
 */
typedef struct _Position {
	PATNode		**node;
	XPFPattern	**pattern;
	XPFFilter	**filter;
} Position;

#define xpfCreatePATNode(m) \
{\
	(m) = (PATNode *)xpfMalloc(sizeof(PATNode));\
	xpfBzero((m), sizeof(PATNode));\
	(m)->info = XPF_INFO_NULL; /* can be different from NULL */\
}

/*
 * Instead of one value in a filter, we store multiple values by putting them
 * in a hash-table. There will only be one such hash-table in a PATNode
 * since there must be a split at that point anyway. 
 * The hashing key is just the value. The hash-table is deleted only when
 * the refcount of each entry is 0 (and the entry is deleted).
 */

typedef PATNode HashEntry;

/*
 * This hashes ALL the first 4 bytes and or more: while inserting, therefore,
 * you must make sure that the other 4 bytes are zeroed out.
 * The type passed must be +ve (-ve types indicate the extractor type for
 * multivalued extractors: stupid but saves a flag! User is not aware of this).
 */
#define xpfHashIndex(pvalue)	( (*(unsigned long *)(pvalue)) % XPF_DAGS_HASHTABLESIZE)

/* Creates hash-table in the extractor and returns the old-value in oldvaluep */
#define xpfCreateHashTable(filterp, oldvaluep) \
{\
	xpfMemcpy(oldvaluep, (filterp)->value, XPF_DATA_MAXSIZE);\
	*(unsigned long *)((filterp)->value) = (unsigned long)(HashEntry **)xpfMalloc(sizeof(HashEntry *) * XPF_DAGS_HASHTABLESIZE);\
	xpfBzero( (HashEntry **)*(unsigned long *)((filterp)->value), sizeof(HashEntry *) * XPF_DAGS_HASHTABLESIZE);\
}

#if	DELAYEDMSGS && STOREMSGS
/* Must take xpfAbs(the actual datatype in the extractor), assign it a local and pass it as 3rd argument */
#define xpfCreateHashEntry(e, valuep, len) \
{\
	e = (HashEntry *)xpfMalloc(sizeof(HashEntry));\
	e->nodetype = XPF_DAGS_HASHENTRY;\
	e->common.hash.alarm = 0;\
	e->common.hash.processnum = 0;\
	e->common.hash.programlist = NULL;\
	xpfMemcpy(e->common.hash.value, valuep, len);\
	if (len < XPF_DATA_MAXSIZE) xpfBzero((char *)(e->common.hash.value + len), XPF_DATA_MAXSIZE - len);\
	e->messagelist = NULL;\
	e->info = XPF_INFO_NULL;\
	e->refcount = 1;\
	e->and = NULL;\
	e->or = NULL;\
}
#else	DELAYEDMSGS && STOREMSGS
/* Must take xpfAbs(the actual datatype in the extractor), assign it a local and pass it as 3rd argument */
#define xpfCreateHashEntry(e, valuep, len) \
{\
	e = (HashEntry *)xpfMalloc(sizeof(HashEntry));\
	e->nodetype = XPF_DAGS_HASHENTRY;\
	e->common.hash.alarm = 0;\
	e->common.hash.processnum = 0;\
	e->common.hash.programlist = NULL;\
	xpfMemcpy(e->common.hash.value, valuep, len);\
	if (len < XPF_DATA_MAXSIZE) xpfBzero((char *)(e->common.hash.value + len), XPF_DATA_MAXSIZE - len);\
	e->info = XPF_INFO_NULL;\
	e->refcount = 1;\
	e->and = NULL;\
	e->or = NULL;\
}
#endif	DELAYEDMSGS && STOREMSGS

/* The index below must already exist due to the find where xpfHashIndex is used */
#define xpfInsertHashEntry(hashtable, e, index) \
{\
	e->or = (hashtable)[index];\
	(hashtable)[index] = e;\
}

/* Written so that *pe can be accessed after deletion but pe doesn't point to anywhere in the DAG */
#define xpfDeleteHashEntry(pe) \
{	/* te is declared in xpfDeletePatternList */\
	te = *pe;\
	*pe = (*pe)->or;\
	pe = &te;\
}

#define xpfDestroyHashEntry(e)	xpfFree((char *)e, sizeof(HashEntry))

/* Frees the hash-table */
#define xpfDestroyHashTable(filterp) \
{\
	xpfFree((*(unsigned long *)(filterp)->value), (sizeof(HashEntry *) * XPF_DAGS_HASHTABLESIZE));\
	(filterp)->refcount = 0;\
}

/*
 * Flag values to indicate where the source and destination patterns are
 * w.r.t. each other. Initially, both will be in different headers (0x11). This
 * does * not say anything directly about headers. Last two are masks.
 */
#define XPF_DAGS_SRCOLDDESTOLD	0x0
#define XPF_DAGS_SRCOLDDESTNEW	0x01
#define XPF_DAGS_SRCISNEW	0x10
#define XPF_DAGS_SRCNEWDESTOLD	0x10
#define XPF_DAGS_SRCNEWDESTNEW	0x11
#define XPF_DAGS_DESTISNEW	0x01

typedef void		(*XPFPutType)();
typedef XPFDataUnit	*(*XPFGetType)();

/*
 * We use the PAT DAG (patricia tree) concept to create an AssociationSet: this
 * is used to store patterns by identifying common-prefixes of their filterlists
 * and merging them together, and creating an associated information list. An
 * incoming message can match many patterns: the job of XPF is to return the
 * info associated with the longest matching (i.e., most specific) pattern.
 *
 * The association set can also be used to trigger "processes" which are
 * generated from user-specified "programs" telling XPF how to extract
 * specific data from incoming messages and form patterns out of them, so that
 * later messages can match patterns other than the ones initially setup.
 * This is useful since the user need not estimate all the possible
 * patterns that could be matched by incoming messages: instead, (s)he can
 * DEFER the creation of new patterns until messages that CAN match such
 * patterns actually arrive.
 *
 * When a process is triggered, the info associated with the generator-
 * pattern is COPIED and then associated with that pattern so that the user
 * can CHANGE the info of that process WITHOUT effecting the info
 * associated with the generator-pattern (THIS should be kept unchanged since
 * many processes can be generated by a pattern and we wouldn't want the
 * info generated for a specific process effect that of another).
 * When MERGEINFO is NOT set, then a COPY of the information is NOT made
 * since as far as XPF knows, it is a simple piece of information.
 *
 * XPF makes sure that the processes generated by a pattern are matched FIRST
 * before the pattern itself is matched.

 * XPF also provides a facility to the user to delay the processing of a
 * message when it does not match a pattern (i.e., instead of asking XPF to pro-
 * ceed when a pattern does NOT match, the user can ask XPF to delay the msg
 * at that pattern and WAIT until a message that actually matches that pattern
 * comes by -- with the hope that a new process that matches the old message
 * might get generated by this new message). Messages can be delayed at nodes
 * with patterns -- not at those with processes.
 *
 * XPF makes sure that the user cannot directly operate on any process-node,
 * i.e., cannot direclty create and destroy processes like patterns:
 * (s)he can do so only implicitly by specifying programs and an alarm "time"
 * after which processes are deleted (see "alarm" in xpf_user.h).
 *
 * Note that a "node" here corr. to a decision, not to protocol-boundaries:
 * infact there might be splits in one protocol/or no splits in many protocols.
 */

typedef struct _XPFState {
	/* list of patterns 1<->1 info */
	PATNode		*xpfAssociationSet;

	/* store the path traversed and multivalued extractor's hashentries */
	PATNode		**xpfMatchedNodes[XPF_DAGS_MAXDEPTH];
	XPFDataUnit	*xpfHeaderData[XPF_DAGS_MAXDEPTH];
	int		xpfHeaderSizes[XPF_DAGS_MAXDEPTH];
	int             xpfMatchedNum;	/* Total number of matched nodes */
	int		xpfHeaderNum;	/* Total number of headers = index+1 */
	int		xpfHeaderCurrent;/* Current header number = index+1 */
	int		xpfHeaderTotalSize;

	/* These are user set/get-able */
	int		xpfMaxAssociations;	/* ignore setups after this */
	int		xpfNumAssociations;	/* current number of ~ */
	int		xpfGranularity;	/* for selecting common prefix */
	int		xpfTimer;	/* frequency of calling Terminate in Launch */
	XPFGetType	(*xpfGetMessageContents)();
	XPFPutType	(*xpfPutMessageContents)();

	int		xpfClock;	/* ++ on each Launch, expires at Timer */
	int		xpfErrno;	/* Special status info, GetPATNode() */
} XPFState;

/*
 * Restore state of current pointers only.
 */
#define xpfRestoreMatch(state) (state)->xpfHeaderCurrent = 0;

/*
 * Restores state of matched nodes and current pointers only.
 */
#define xpfRestoreNoMatch(state) \
{\
	xpfRestoreMatch(state);\
	(state)->xpfMatchedNum = 0;\
}

/*
 * Restores state of extracted headers, matched nodes and current pointers only.
 */
#define xpfRestoreHeaders(state) \
{\
	xpfRestoreNoMatch(state);\
	(state)->xpfHeaderNum = 0;\
	(state)->xpfHeaderTotalSize = 0;\
}

/*
 * Restores the state of the message, extracted headers, matched nodes and current pointers.
 */
#define xpfRestoreAll(state, message, length) \
{\
	int	_i;\
\
	if ((state)->xpfPutMessageContents != NULL) \
		for (_i=(state)->xpfHeaderNum; _i>0; _i--) {	/* Put in reverse order of Get */\
			((state)->xpfPutMessageContents)(message, length, (state)->xpfHeaderTotalSize, (state)->xpfHeaderSizes[_i-1], (state)->xpfHeaderData[_i-1]);\
		}\
	xpfRestoreHeaders(state);\
}

#define xpfAssert(cond)\
{\
	if (!(cond)) {\
		xpfPrintf("XPF: Assertion failed @ %s:%d!\n", __FILE__, __LINE__);\
	} /* Do not exit since code resorts to default operations/returns error */\
}

#if	VARHEADER
#define COPY_HEADER(hdr1, hdr2) \
{	/* NOTE: if hdr1 was valid, the old "currency" information is now lost, as it should be */\
	xpfMemcpy((hdr1), (hdr2), sizeof(XPFHeader));\
	if ((hdr2)->headertype & XPF_HEADER_VARIABLE) {\
		if ((hdr1)->size.variable == NULL) (hdr1)->size.variable = (XPFVariableHeader *)xpfMalloc(sizeof(XPFVariableHeader));\
		xpfMemcpy((hdr1)->size.variable, (hdr2)->size.variable, sizeof(XPFVariableHeader));\
	}\
}
#else	VARHEADER
#define COPY_HEADER(hdr1, hdr2) \
{	/* NOTE: if hdr1 was valid, the old "currency" information is now lost, as it should be */\
	xpfMemcpy((hdr1), (hdr2), sizeof(XPFHeader));\
}
#endif	VARHEADER

/*
 * The message is simply interpreted as a buffer (XPFDataUnit *) pointer
 * if get/put functions are not defined. Else they are called.
 * This is why the xpfHeader.* state is maintained in by xpf.
 */

#define GET_CONTENTS(functionptr, message, length, presentsize, neededsize) \
	(((presentsize + neededsize) > length) ? NULL :\
		(((functionptr) == NULL) ? (XPFDataUnit *)(message + presentsize) :\
			(XPFDataUnit *)(*functionptr)(message, length, presentsize, neededsize)))

#define PUT_CONTENTS(functionptr, message, length, presentsize, givensize, buffer) \
	if ((functionptr) != NULL) (*(functionptr))(message, length, presentsize, givensize, buffer);

/*
 * Macro to extract a header from "message" and store it in a buffer.
 * If any header has already been extracted once, it saves time by NOT calling the
 * GetMessageContents() routine again. The logic is that each search might pass
 * thru different filters but the protocols in the message remain as they are.
 * A PutMessageContents() is done ONLY before Demuxing the message upward.
 */
#if	VARHEADER
#define xpfExtractHeader(state, message, length, hdr, ret)\
{\
	unsigned int	offset;\
	XPFDataUnit	*buffer;\
\
/* SAFETY\
	if (hdr == NULL) {\
		ret = FALSE;\
		goto hdr_done;\
	}\
*/\
	DOPRINT(xpfPrintf("In xpfExtractHdr: size = %x\n", hdr->size);)\
	if (hdr->headertype & XPF_HEADER_CURRENT) {	/* part of the same protocol as before */\
		ret = TRUE;\
		goto hdr_done;\
	}\
\
	if (state->xpfHeaderNum > state->xpfHeaderCurrent) {	/* second run of matching: headers will be same, the filters will be different */\
		DOPRINT(xpfPrintf("Extract: old %x\n", state->xpfHeaderCurrent + 1);)\
		state->xpfHeaderCurrent ++;\
		ret = TRUE;\
		goto hdr_done;\
	}\
	else if (state->xpfHeaderCurrent >= XPF_DAGS_MAXDEPTH) {\
		DOPRINT(xpfPrintf("Extract error: depth about to exceed max-dag-depth %x\n", XPF_DAGS_MAXDEPTH);)\
		ret = FALSE;\
		goto hdr_done;\
	}\
\
	DOPRINT(xpfAssert(state->xpfHeaderCurrent == state->xpfHeaderNum));	/* must go in lock-step */\
\
	if (hdr->headertype & XPF_HEADER_VARIABLE) {	/* variable length headers */\
		offset = hdr->size.variable->extractor.selector.offset XPF_USER_GETCOUNTUNIT(hdr->size.variable);\
		switch(hdr->size.variable->extractor.datatype)\
		{\
		case XPF_DATA_CHAR:\
		{\
			char	temp;\
\
			if ((buffer = GET_CONTENTS(state->xpfGetMessageContents, message, length, state->xpfHeaderTotalSize,\
							offset + 1)) == NULL) {\
				ret = FALSE;\
				goto hdr_done;\
			}\
			temp = buffer[offset] & hdr->size.variable->extractor.mask[0];\
			if (hdr->size.variable->multiplier >= 0) temp *= hdr->size.variable->multiplier;\
			else temp /= (-hdr->size.variable->multiplier);\
\
			/* the field which has the size of the header is at the end of the header */\
			if (temp <= offset) temp = offset;\
			else {\
				PUT_CONTENTS(state->xpfPutMessageContents, message, length, state->xpfHeaderTotalSize +\
						offset+1, offset+1, buffer);\
				if ((buffer=GET_CONTENTS(state->xpfGetMessageContents, message, length, state->xpfHeaderTotalSize, temp)) == NULL) {\
					ret = FALSE;\
					goto hdr_done;\
				}\
				offset = temp;\
			}\
			break;\
		}\
\
		case XPF_DATA_SHORT:\
		{\
			short	temp;\
\
			if ((buffer = GET_CONTENTS(state->xpfGetMessageContents, message, length, state->xpfHeaderTotalSize, offset + 2)) == NULL) {\
				DOPRINT(xpfPrintf("Extract error: want=%x curr=%x\n", offset + 2, state->xpfHeaderTotalSize);)\
				ret = FALSE;\
				goto hdr_done;\
			}\
			((char *)&temp)[0] = buffer[offset] & hdr->size.variable->extractor.mask[0];\
			((char *)&temp)[1] = buffer[offset + 1] & hdr->size.variable->extractor.mask[1];\
			if (hdr->size.variable->multiplier >= 0) temp *= hdr->size.variable->multiplier;\
			else temp /= (-hdr->size.variable->multiplier);\
			if (temp <= offset) temp = offset;	/* the field which has the size of the header is at the end of the header */\
			else {\
				PUT_CONTENTS(state->xpfPutMessageContents, message, length, state->xpfHeaderTotalSize + offset+2, offset+2, buffer);\
				if ((buffer=GET_CONTENTS(state->xpfGetMessageContents, message, length, state->xpfHeaderTotalSize, temp)) == NULL) {\
					DOPRINT(xpfPrintf("Extract error: want=%x curr=%x\n", temp-offset-2, state->xpfHeaderTotalSize + offset + 2);)\
					ret = FALSE;\
					goto hdr_done;\
				}\
				offset = temp;\
			}\
			break;\
		}\
\
		case XPF_DATA_INT:\
		{\
			int	temp;\
\
			if ((buffer = GET_CONTENTS(state->xpfGetMessageContents, message, length, state->xpfHeaderTotalSize, offset + sizeof(int))) == NULL) {\
				DOPRINT(xpfPrintf("Extract error: want=%x curr=%x\n", offset + sizeof(int), state->xpfHeaderTotalSize);)\
				ret = FALSE;\
				goto hdr_done;\
			}\
			if ( (unsigned long)(buffer + offset) & 0x3) {\
				((char *)&temp)[0] = buffer[offset] & hdr->size.variable->extractor.mask[0];\
				((char *)&temp)[1] = buffer[offset + 1] & hdr->size.variable->extractor.mask[1];\
				((char *)&temp)[2] = buffer[offset + 2] & hdr->size.variable->extractor.mask[2];\
				((char *)&temp)[3] = buffer[offset + 3] & hdr->size.variable->extractor.mask[3];\
			}\
			else temp = *((int *)&(buffer[offset])) & *((int *)&(hdr->size.variable->extractor.mask[0]));\
			if (hdr->size.variable->multiplier >= 0) temp *= hdr->size.variable->multiplier;\
			else temp /= (-hdr->size.variable->multiplier);\
			if (temp <= offset) temp = offset;	/* the field which has the size of the header is at the end of the header */\
			else {\
				PUT_CONTENTS(state->xpfPutMessageContents, message, length, state->xpfHeaderTotalSize + offset+sizeof(int), offset+sizeof(int), buffer);\
				if ((buffer=GET_CONTENTS(state->xpfGetMessageContents, message, length, state->xpfHeaderTotalSize, temp)) == NULL) {\
					DOPRINT(xpfPrintf("Extract error: want=%x curr=%x\n", temp-offset-sizeof(int), state->xpfHeaderTotalSize + offset + sizeof(int));)\
					ret = FALSE;\
					goto hdr_done;\
				}\
				offset = temp;\
			}\
			break;\
		}\
\
		case XPF_DATA_NONE:\
		default:\
			ret = FALSE;\
			goto hdr_done;\
		}\
\
		/* offset is the total number of bytes to be extracted from the packet */\
		state->xpfHeaderTotalSize += offset;\
		state->xpfHeaderSizes[state->xpfHeaderCurrent] = offset;\
		state->xpfHeaderData[state->xpfHeaderCurrent] = buffer;\
		state->xpfHeaderCurrent = ++state->xpfHeaderNum;\
		DOPRINT(xpfPrintf("Extract: newnum %x newbytes %x\n", state->xpfHeaderCurrent, state->xpfHeaderTotalSize);)\
	}\
	else {	/* fixed length headers */\
		if ((state->xpfHeaderData[state->xpfHeaderCurrent] =\
		     GET_CONTENTS(state->xpfGetMessageContents, message, length, state->xpfHeaderTotalSize, hdr->size.fixed)) == NULL) {\
			ret = FALSE;\
			goto hdr_done;\
		}\
		state->xpfHeaderTotalSize += hdr->size.fixed;\
		state->xpfHeaderSizes[state->xpfHeaderCurrent] = hdr->size.fixed;\
		state->xpfHeaderCurrent = ++state->xpfHeaderNum;\
		DOPRINT(xpfPrintf("Extract: newnum %x newbytes %x\n", state->xpfHeaderCurrent, state->xpfHeaderTotalSize);)\
	}\
	ret = TRUE;\
hdr_done:\
	;\
}
#else	VARHEADER
#define xpfExtractHeader(state, message, length, hdr, ret)\
{\
/* SAFETY\
	if (hdr == NULL) return FALSE;\
*/\
	DOPRINT(xpfPrintf("In xpfExtractHdr: size = %x\n", hdr->size);)\
	if (hdr->headertype & XPF_HEADER_CURRENT) {	/* part of the same protocol as before */\
		ret = TRUE;\
		goto hdr_done;\
	}\
\
	if (state->xpfHeaderNum > state->xpfHeaderCurrent) {	/* second run of matching: headers will be same, the filters will be different */\
		DOPRINT(xpfPrintf("Extract: old %x\n", state->xpfHeaderCurrent + 1);)\
		state->xpfHeaderCurrent ++;\
		ret = TRUE;\
		goto hdr_done;\
	}\
	else if (state->xpfHeaderCurrent >= XPF_DAGS_MAXDEPTH) {\
		DOPRINT(xpfPrintf("Extract error: depth about to exceed max-dag-depth %x\n", XPF_DAGS_MAXDEPTH);)\
		ret = FALSE;\
		goto hdr_done;\
	}\
\
	DOPRINT(xpfAssert(state->xpfHeaderCurrent == state->xpfHeaderNum));	/* must go in lock-step */\
\
		if ((state->xpfHeaderData[state->xpfHeaderCurrent] =\
		     GET_CONTENTS(state->xpfGetMessageContents, message, length, state->xpfHeaderTotalSize, hdr->size.fixed)) == NULL) {\
			ret = FALSE;\
			goto hdr_done;\
		}\
		state->xpfHeaderTotalSize += hdr->size.fixed;\
		state->xpfHeaderSizes[state->xpfHeaderCurrent] = hdr->size.fixed;\
		state->xpfHeaderCurrent = ++state->xpfHeaderNum;\
		DOPRINT(xpfPrintf("Extract: newnum %x newbytes %x\n", state->xpfHeaderCurrent, state->xpfHeaderTotalSize);)\
\
	ret = TRUE;\
hdr_done:\
	;\
}
#endif	VARHEADER

/*
 * This masks buffer[offset] with the mask in the current `filter' to obtain
 * a value, and compares this value with the value in filter and returns <, >, =.
 * If there is an error in generating the offset, then it returns XPF_DAGS_ERROR.
 * If the type is XPF_DATA_NONE, since the filter matches all packets, it returns XPF_DAGS_SAME.
 *
 * We don't have to return XPF_DAGS_BEFORE or XPF_DAGS_AFTER since this is never used to
 * compare filters: only to compare a messsage with a filter => we take the NO
 * branch during search anyway if it is NOT XPF_DAGS_SAME: we can just return XPF_DAGS_AFTER.
 * "pe" is set to the matching hash-table-value if any only in the -ve type case:
 * otherwise, it is untouched (this function is called with *pe set to NULL).
 */

#if	STRINGTYPE && BYTEADDRESS
#define EXTRACT_STRINGTYPE\
		case XPF_DATA_STRING:\
			if ((offset < 0) || (offset >= maxsize)) {\
				ret = XPF_DAGS_ERROR;	/* stringlen must be @least 1 */\
				goto end_compare;\
			}\
			if ((ret = strnmaskordercmp(&(buffer[offset]),\
							filter->value,\
							filter->extractor.mask,\
							(XPF_DATA_MAXSIZE > maxsize - offset) ? (maxsize - offset) : XPF_DATA_MAXSIZE)) != 0) {\
				ret =  XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			else break;
#define EXTRACT_STRINGTYPE_HASH\
		case XPF_DATA_STRING:\
			if ((offset < 0) || (offset >= maxsize)) {\
				ret = XPF_DAGS_ERROR;	/* stringlen must be @least 1 */\
				goto end_compare;\
			}\
			strnmaskcpy(pvalue, &buffer[offset], filter->extractor.mask, (XPF_DATA_MAXSIZE > maxsize - offset) ? (maxsize - offset) : XPF_DATA_MAXSIZE);\
			size = XPF_DATA_MAXSIZE;\
			break;
#else	STRINGTYPE && BYTEADDRESS
#define EXTRACT_STRINGTYPE
#define EXTRACT_STRINGTYPE_HASH
#endif	STRINGTYPE && BYTEADDRESS

#if	BYTEADDRESS
#if	AVOIDBUSERROR
#define xpfCompareWithFilter(state, filter, pe, ret)\
{\
	int		offset;\
	int		type;\
	int		index;\
	XPFDataUnit	pvalue[XPF_DATA_MAXSIZE];\
	unsigned char	*buffer;\
	int		maxsize;\
	int		level;\
\
	level = filter->extractor.selector.level;\
	if (level == 0) {	/* look at your own header */\
		buffer = state->xpfHeaderData[state->xpfHeaderCurrent-1];\
		maxsize = state->xpfHeaderSizes[state->xpfHeaderCurrent-1];\
	}\
	else if (level > 0) {	/* peek behind into protocols below you in stack */\
		if (level > state->xpfHeaderCurrent-1) {\
			ret = XPF_DAGS_ERROR;\
			goto end_compare;\
		}\
		buffer = state->xpfHeaderData[state->xpfHeaderCurrent-1-level];\
		maxsize = state->xpfHeaderSizes[state->xpfHeaderCurrent-1-level];\
	}\
	else {	/* peek ahead into protocols above you in stack */\
		ret = XPF_DAGS_ERROR;\
		goto end_compare;\
		/*\
		maxsize = state->xpfCurrentMessageSize;\
		buffer = how to look into the new buffer without access to the next protocol's header node?\
		You still cannot overlapp headers since that makes no sense (each protocol puts its headers independently).\
		*/\
	}\
\
/*\
	if ((buffer == NULL) || (maxsize <= 0) || (filter == NULL))\
		ret = XPF_DAGS_ERROR;\
		goto end_compare;\
*/\
	if ((type = filter->extractor.datatype) > 0) {\
		offset = filter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(filter);\
		NOPRINT(xpfPrintf("Compare: off %x type %x\n", offset, type);)\
\
		switch(type)\
		{\
		case XPF_DATA_NONE: break;\
\
		case XPF_DATA_CHAR:\
			if ((offset < 0) || (offset >= maxsize)) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if( (buffer[offset] & filter->extractor.mask[0]) != filter->value[0]) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			break;\
\
		case XPF_DATA_SHORT:\
			if ((offset < 0) || (offset >= maxsize + 1 - sizeof(short))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ((unsigned long)(buffer+offset) % sizeof(short)) {\
				type = sizeof(short); goto avoid_buserror1;\
			}\
			if ( ((*((short *)&(buffer[offset]))) & *((short *)&(filter->extractor.mask[0]))) !=\
			     *((short *)&(filter->value[0])) ) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			break;\
\
		case XPF_DATA_INT:\
			if ((offset < 0) || (offset>= maxsize + 1 - sizeof(int))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ((unsigned long)(buffer+offset) % sizeof(int)) {\
				type = sizeof(int); goto avoid_buserror1;\
			}\
			if ( ((*((int *)&(buffer[offset]))) & *((int *)&(filter->extractor.mask[0]))) !=\
			     *((int *)&(filter->value[0])) ) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			break;\
\
		EXTRACT_STRINGTYPE;\
\
		default: /* type is length: type <= XPF_DATA_MAXSIZE if (not one of above) is checked at filter-setup time */\
			type = xpfSizeFromType(type);\
		avoid_buserror1:\
			if ( (offset < 0) || ((type + offset) >= (maxsize + 1)) ) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ((ret = bmaskordercmp(&(buffer[offset]),\
						filter->value,\
						filter->extractor.mask,\
						type)) != 0) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			else break;\
		} /* switch */\
	}\
	else {	/* Must extract the pvalue from the buffer, and THEN try to hash it in the hash-table at the extractor: can't compare directly as above */\
		int		size;\
		HashEntry	**hashtable;\
		int		j;\
\
		if ((type = -type) == XPF_DATA_NONE) {\
			ret = XPF_DAGS_ERROR;	/* multivalued extractor can't matching everything: xpfInsertPatternList() */\
			goto end_compare;\
		}\
		offset = filter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(filter);\
		NOPRINT(xpfPrintf("Compare: off %x type %x\n", offset, type);)\
		xpfBzero(pvalue, XPF_DATA_MAXSIZE);\
\
		switch(type)\
		{\
		case XPF_DATA_CHAR:\
			if ((offset < 0) || (offset >= maxsize)) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			pvalue[0] = buffer[offset] & filter->extractor.mask[0];\
			size = sizeof(char);\
			break;\
\
		case XPF_DATA_SHORT:\
			if ((offset < 0) || (offset >= maxsize + 1 - sizeof(short))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ((unsigned long)(buffer+offset) % sizeof(short)) {\
				size = type = sizeof(short); goto avoid_buserror2;\
			}\
			*((short *)&(pvalue[0])) = ((*((short *)&(buffer[offset]))) & *((short *)&(filter->extractor.mask[0])));\
			size = sizeof(short);\
			break;\
\
		case XPF_DATA_INT:\
			if ((offset < 0) || (offset>= maxsize + 1 - sizeof(int))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ((unsigned long)(buffer+offset) % sizeof(int)) {\
				size = type = sizeof(int); goto avoid_buserror2;\
			}\
			*((int *)&(pvalue[0])) = ((*((int *)&(buffer[offset]))) & *((int *)&(filter->extractor.mask[0])));\
			size = sizeof(int);\
			break;\
\
		EXTRACT_STRINGTYPE_HASH\
\
		default: /* type is length: type <= XPF_DATA_MAXSIZE if (not one of above) is checked at filter-setup time */\
			size = type = xpfSizeFromType(type);\
		avoid_buserror2:\
			if ( (offset < 0) || ((type + offset) >= (maxsize + 1)) ) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			memmaskcpy(pvalue, &buffer[offset], filter->extractor.mask, type);\
			if ((ret = bmaskordercmp(&(buffer[offset]),\
						filter->value,\
						filter->extractor.mask,\
						type)) != 0) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			else break;\
		} /* switch */\
\
		/* remove function calls: make inline code: hopefully, reduce single->multi valued timing jump */\
		index = xpfHashIndex(pvalue);\
\
		if ((hashtable = (HashEntry **)*(unsigned long *)(filter->value)) == NULL) {\
			ret = XPF_DAGS_AFTER;\
			goto end_compare;\
		}\
		pe = &hashtable[index];\
		while (*pe != NULL) {\
			/* LLP's suggestion: x-Kernel fundas about fixed size hash-keys\
			for (j=0; j<size; j++)\
				if ((*pe)->common.hash.value[j] != pvalue[j]) break;\
			if (j>=size) break;\
			*/\
			if ( *((unsigned long *)&(pvalue[0])) == *((unsigned long *)(*pe)->common.hash.value) ) break;\
			pe = &(*pe)->or;\
		}\
\
		if (*pe == NULL) {\
			ret = XPF_DAGS_AFTER;\
			goto end_compare;\
		}\
	}\
	ret = XPF_DAGS_SAME;\
end_compare:\
	;\
}
#else	AVOIDBUSERROR
/*
 * This is the default
 */
#define xpfCompareWithFilter(state, filter, pe, ret)\
{\
	int		offset;\
	int		type;\
	int		index;\
	XPFDataUnit	pvalue[XPF_DATA_MAXSIZE];\
	unsigned char	*buffer;\
	int		maxsize;\
	int		level;\
\
	level = filter->extractor.selector.level;\
	if (level == 0) {	/* look at your own header */\
		buffer = state->xpfHeaderData[state->xpfHeaderCurrent-1];\
		maxsize = state->xpfHeaderSizes[state->xpfHeaderCurrent-1];\
	}\
	else if (level > 0) {	/* peek behind into protocols below you in stack */\
		if (level > state->xpfHeaderCurrent-1) {\
			ret = XPF_DAGS_ERROR;\
			goto end_compare;\
		}\
		buffer = state->xpfHeaderData[state->xpfHeaderCurrent-1-level];\
		maxsize = state->xpfHeaderSizes[state->xpfHeaderCurrent-1-level];\
	}\
	else {	/* peek ahead into protocols above you in stack */\
		ret = XPF_DAGS_ERROR;\
		goto end_compare;\
		/*\
		maxsize = state->xpfCurrentMessageSize;\
		buffer = how to look into the new buffer without access to the next protocol's header node?\
		You still cannot overlapp headers since that makes no sense (each protocol puts its headers independently).\
		*/\
	}\
\
/*\
	if ((buffer == NULL) || (maxsize <= 0) || (filter == NULL))\
		ret = XPF_DAGS_ERROR;\
		goto end_compare;\
*/\
	if ((type = filter->extractor.datatype) > 0) {\
		offset = filter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(filter);\
		NOPRINT(xpfPrintf("Compare: off %x type %x\n", offset, type);)\
\
		switch(type)\
		{\
		case XPF_DATA_NONE: break;\
\
		case XPF_DATA_CHAR:\
			if ((offset < 0) || (offset >= maxsize)) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if( (buffer[offset] & filter->extractor.mask[0]) != filter->value[0]) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			break;\
\
		case XPF_DATA_SHORT:\
			if ((offset < 0) || (offset >= maxsize + 1 - sizeof(short))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ( ((*((short *)&(buffer[offset]))) & *((short *)&(filter->extractor.mask[0]))) !=\
			     *((short *)&(filter->value[0])) ) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			break;\
\
		case XPF_DATA_INT:\
			if ((offset < 0) || (offset>= maxsize + 1 - sizeof(int))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ( ((*((int *)&(buffer[offset]))) & *((int *)&(filter->extractor.mask[0]))) !=\
			     *((int *)&(filter->value[0])) ) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			break;\
\
		EXTRACT_STRINGTYPE;\
\
		default: /* type is length: type <= XPF_DATA_MAXSIZE if (not one of above) is checked at filter-setup time */\
			type = xpfSizeFromType(type);\
		avoid_buserror1:\
			if ( (offset < 0) || ((type + offset) >= (maxsize + 1)) ) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ((ret = bmaskordercmp(&(buffer[offset]),\
						filter->value,\
						filter->extractor.mask,\
						type)) != 0) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			else break;\
		} /* switch */\
	}\
	else {	/* Must extract the pvalue from the buffer, and THEN try to hash it in the hash-table at the extractor: can't compare directly as above */\
		int		size;\
		HashEntry	**hashtable;\
		int		j;\
\
		if ((type = -type) == XPF_DATA_NONE) {\
			ret = XPF_DAGS_ERROR;	/* multivalued extractor can't matching everything: xpfInsertPatternList() */\
			goto end_compare;\
		}\
		offset = filter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(filter);\
		NOPRINT(xpfPrintf("Compare: off %x type %x\n", offset, type);)\
		xpfBzero(pvalue, XPF_DATA_MAXSIZE);\
\
		switch(type)\
		{\
		case XPF_DATA_CHAR:\
			if ((offset < 0) || (offset >= maxsize)) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			pvalue[0] = buffer[offset] & filter->extractor.mask[0];\
			size = sizeof(char);\
			break;\
\
		case XPF_DATA_SHORT:\
			if ((offset < 0) || (offset >= maxsize + 1 - sizeof(short))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			*((short *)&(pvalue[0])) = ((*((short *)&(buffer[offset]))) & *((short *)&(filter->extractor.mask[0])));\
			size = sizeof(short);\
			break;\
\
		case XPF_DATA_INT:\
			if ((offset < 0) || (offset>= maxsize + 1 - sizeof(int))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			*((int *)&(pvalue[0])) = ((*((int *)&(buffer[offset]))) & *((int *)&(filter->extractor.mask[0])));\
			size = sizeof(int);\
			break;\
\
		EXTRACT_STRINGTYPE_HASH\
\
		default: /* type is length: type <= XPF_DATA_MAXSIZE if (not one of above) is checked at filter-setup time */\
			size = type = xpfSizeFromType(type);\
		avoid_buserror2:\
			if ( (offset < 0) || ((type + offset) >= (maxsize + 1)) ) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			memmaskcpy(pvalue, &buffer[offset], filter->extractor.mask, type);\
			if ((ret = bmaskordercmp(&(buffer[offset]),\
						filter->value,\
						filter->extractor.mask,\
						type)) != 0) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			else break;\
		} /* switch */\
\
		/* remove function calls: make inline code: hopefully, reduce single->multi valued timing jump */\
		index = xpfHashIndex(pvalue);\
\
		if ((hashtable = (HashEntry **)*(unsigned long *)(filter->value)) == NULL) {\
			ret = XPF_DAGS_AFTER;\
			goto end_compare;\
		}\
		pe = &hashtable[index];\
		while (*pe != NULL) {\
			/* LLP's suggestion: x-Kernel fundas about fixed size hash-keys\
			for (j=0; j<size; j++)\
				if ((*pe)->common.hash.value[j] != pvalue[j]) break;\
			if (j>=size) break;\
			*/\
			if ( *((unsigned long *)&(pvalue[0])) == *((unsigned long *)(*pe)->common.hash.value) ) break;\
			pe = &(*pe)->or;\
		}\
\
		if (*pe == NULL) {\
			ret = XPF_DAGS_AFTER;\
			goto end_compare;\
		}\
	}\
	ret = XPF_DAGS_SAME;\
end_compare:\
	;\
}
#endif	AVOIDBUSERROR
#else	BYTEADDRESS
#if	AVOIDBUSERROR
#define xpfCompareWithFilter(state, filter, pe, ret)\
{\
	int		offset;\
	int		type;\
	int		index;\
	XPFDataUnit	pvalue[XPF_DATA_MAXSIZE];\
	unsigned char	*buffer;\
	int		maxsize;\
	int		level;\
\
	level = filter->extractor.selector.level;\
	if (level == 0) {	/* look at your own header */\
		buffer = state->xpfHeaderData[state->xpfHeaderCurrent-1];\
		maxsize = state->xpfHeaderSizes[state->xpfHeaderCurrent-1];\
	}\
	else if (level > 0) {	/* peek behind into protocols below you in stack */\
		if (level > state->xpfHeaderCurrent-1) {\
			ret = XPF_DAGS_ERROR;\
			goto end_compare;\
		}\
		buffer = state->xpfHeaderData[state->xpfHeaderCurrent-1-level];\
		maxsize = state->xpfHeaderSizes[state->xpfHeaderCurrent-1-level];\
	}\
	else {	/* peek ahead into protocols above you in stack */\
		ret = XPF_DAGS_ERROR;\
		goto end_compare;\
		/*\
		maxsize = state->xpfCurrentMessageSize;\
		buffer = how to look into the new buffer w/o access to the next protocol's header node (i.e., w/o extracting its header)?\
		You still cannot overlapp headers since that makes no sense (each protocol puts its headers independently).\
		*/\
	}\
\
/*\
	if ((buffer == NULL) || (maxsize <= 0) || (filter == NULL))\
		ret = XPF_DAGS_ERROR;\
		goto end_compare;\
*/\
	if ((type = filter->extractor.datatype) > 0) {\
		offset = filter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(filter);\
		NOPRINT(xpfPrintf("Compare: off %x type %x\n", offset, type);)\
\
		switch(type)\
		{\
		case XPF_DATA_NONE: break;\
\
		case XPF_DATA_CHAR:\
			if ((offset < 0) || (offset >= maxsize)) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if (offset & 0x1)\
				ret = *((short *)&buffer[offset-1]) & 0x00ff;\
			else {\
				if (maxsize-offset<2) {\
					ret = XPF_DAGS_ERROR;	/* must be able to access a short from around offset */\
					goto end_compare;\
				}\
				ret = (*((short *)&buffer[offset]) & 0xff00) >> 8;\
			}\
			if ((ret & filter->extractor.mask[0]) != filter->value[0]) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			break;\
\
		case XPF_DATA_SHORT:\
			if ((offset < 0) || (offset >= maxsize + 1 - sizeof(short))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ((unsigned long)(buffer+offset) % sizeof(short)) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ( ((*((short *)&(buffer[offset]))) & *((short *)&(filter->extractor.mask[0]))) !=\
			     *((short *)&(filter->value[0])) ) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			break;\
\
		case XPF_DATA_INT:\
			if ((offset < 0) || (offset>= maxsize + 1 - sizeof(int))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ((unsigned long)(buffer+offset) % sizeof(int)) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ( ((*((int *)&(buffer[offset]))) & *((int *)&(filter->extractor.mask[0]))) !=\
			     *((int *)&(filter->value[0])) ) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			break;\
\
		EXTRACT_STRINGTYPE;\
\
		default: /* type is length: type <= XPF_DATA_MAXSIZE if (not one of above) is checked at filter-setup time */\
			type = xpfSizeFromType(type);\
			if (!(offset & 0x1) && (type & 0x1)) {\
				ret = XPF_DAGS_ERROR;	/* must be able to access a short around each offset */\
				goto end_compare;\
			}\
		avoid_buserror1:\
			if ( (offset < 0) || ((type + offset) >= (maxsize + 1)) ) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ((ret = bmaskordercmp(&(buffer[offset]),\
						filter->value,\
						filter->extractor.mask,\
						type)) != 0) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			else break;\
		} /* switch */\
	}\
	else {	/* Must extract the pvalue from the buffer, and THEN try to hash it in the hash-table at the extractor: can't compare directly as above */\
		int		size;\
		HashEntry	**hashtable;\
		int		j;\
\
		if ((type = -type) == XPF_DATA_NONE) {\
			ret = XPF_DAGS_ERROR;	/* multivalued extractor can't matching everything: xpfInsertPatternList() */\
			goto end_compare;\
		}\
		offset = filter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(filter);\
		NOPRINT(xpfPrintf("Compare: off %x type %x\n", offset, type);)\
		xpfBzero(pvalue, XPF_DATA_MAXSIZE);\
\
		switch(type)\
		{\
		case XPF_DATA_CHAR:\
			if ((offset < 0) || (offset >= maxsize)) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if (offset&0x1)\
				pvalue[0] = *((short *)&buffer[offset-1]) & 0x00ff;\
			else {\
				if (maxsize-offset<2) {\
					ret = XPF_DAGS_ERROR;	/* must be able to access a short from around offset */\
					goto end_compare;\
				}\
				pvalue[0] = (*((short *)&buffer[offset]) & 0xff00) >> 8;\
			}\
			size = sizeof(char);\
			break;\
\
		case XPF_DATA_SHORT:\
			if ((offset < 0) || (offset >= maxsize + 1 - sizeof(short))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ((unsigned long)(buffer+offset) % sizeof(short)) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			*((short *)&(pvalue[0])) = ((*((short *)&(buffer[offset]))) & *((short *)&(filter->extractor.mask[0])));\
			size = sizeof(short);\
			break;\
\
		case XPF_DATA_INT:\
			if ((offset < 0) || (offset>= maxsize + 1 - sizeof(int))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ((unsigned long)(buffer+offset) % sizeof(int)) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			*((int *)&(pvalue[0])) = ((*((int *)&(buffer[offset]))) & *((int *)&(filter->extractor.mask[0])));\
			size = sizeof(int);\
			break;\
\
		EXTRACT_STRINGTYPE_HASH\
\
		default: /* type is length: type <= XPF_DATA_MAXSIZE if (not one of above) is checked at filter-setup time */\
			size = type = xpfSizeFromType(type);\
			if (!(offset & 0x1) && (type & 0x1)) {\
				ret = XPF_DAGS_ERROR;	/* must be able to access a short around each offset */\
				goto end_compare;\
			}\
		avoid_buserror2:\
			if ( (offset < 0) || ((type + offset) >= (maxsize + 1)) ) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			memmaskcpy(pvalue, &buffer[offset], filter->extractor.mask, type);\
			if ((ret = bmaskordercmp(&(buffer[offset]),\
						filter->value,\
						filter->extractor.mask,\
						type)) != 0) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			else break;\
		} /* switch */\
\
		/* remove function calls: make inline code: hopefully, reduce single->multi valued timing jump */\
		index = xpfHashIndex(pvalue);\
\
		if ((hashtable = (HashEntry **)*(unsigned long *)(filter->value)) == NULL) {\
			ret = XPF_DAGS_AFTER;\
			goto end_compare;\
		}\
		pe = &hashtable[index];\
		while (*pe != NULL) {\
			/* LLP's suggestion: x-Kernel fundas about fixed size hash-keys\
			for (j=0; j<size; j++)\
				if ((*pe)->common.hash.value[j] != pvalue[j]) break;\
			if (j>=size) break;\
			*/\
			if ( *((unsigned long *)&(pvalue[0])) == *((unsigned long *)(*pe)->common.hash.value) ) break;\
			pe = &(*pe)->or;\
		}\
\
		if (*pe == NULL) {\
			ret = XPF_DAGS_AFTER;\
			goto end_compare;\
		}\
	}\
	ret = XPF_DAGS_SAME;\
end_compare:\
	;\
}
#else	AVOIDBUSERROR
#define xpfCompareWithFilter(state, filter, pe, ret)\
{\
	int		offset;\
	int		type;\
	int		index;\
	XPFDataUnit	pvalue[XPF_DATA_MAXSIZE];\
	unsigned char	*buffer;\
	int		maxsize;\
	int		level;\
\
	level = filter->extractor.selector.level;\
	if (level == 0) {	/* look at your own header */\
		buffer = state->xpfHeaderData[state->xpfHeaderCurrent-1];\
		maxsize = state->xpfHeaderSizes[state->xpfHeaderCurrent-1];\
	}\
	else if (level > 0) {	/* peek behind into protocols below you in stack */\
		if (level > state->xpfHeaderCurrent-1) {\
			ret = XPF_DAGS_ERROR;\
			goto end_compare;\
		}\
		buffer = state->xpfHeaderData[state->xpfHeaderCurrent-1-level];\
		maxsize = state->xpfHeaderSizes[state->xpfHeaderCurrent-1-level];\
	}\
	else {	/* peek ahead into protocols above you in stack */\
		ret = XPF_DAGS_ERROR;\
		goto end_compare;\
		/*\
		maxsize = state->xpfCurrentMessageSize;\
		buffer = how to look into the new buffer without access to the next protocol's header node?\
		You still cannot overlapp headers since that makes no sense (each protocol puts its headers independently).\
		*/\
	}\
\
/*\
	if ((buffer == NULL) || (maxsize <= 0) || (filter == NULL))\
		ret = XPF_DAGS_ERROR;\
		goto end_compare;\
*/\
	if ((type = filter->extractor.datatype) > 0) {\
		offset = filter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(filter);\
		NOPRINT(xpfPrintf("Compare: off %x type %x\n", offset, type);)\
\
		switch(type)\
		{\
		case XPF_DATA_NONE: break;\
\
		case XPF_DATA_CHAR:\
			if ((offset < 0) || (offset >= maxsize)) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if (offset & 0x1)\
				ret = *((short *)&buffer[offset-1]) & 0x00ff;\
			else {\
				if (maxsize-offset<2) {\
					ret = XPF_DAGS_ERROR;	/* must be able to access a short from around offset */\
					goto end_compare;\
				}\
				ret = (*((short *)&buffer[offset]) & 0xff00) >> 8;\
			}\
			if ((ret & filter->extractor.mask[0]) != filter->value[0]) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			break;\
\
		case XPF_DATA_SHORT:\
			if ((offset < 0) || (offset >= maxsize + 1 - sizeof(short))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ( ((*((short *)&(buffer[offset]))) & *((short *)&(filter->extractor.mask[0]))) !=\
			     *((short *)&(filter->value[0])) ) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			break;\
\
		case XPF_DATA_INT:\
			if ((offset < 0) || (offset>= maxsize + 1 - sizeof(int))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ( ((*((int *)&(buffer[offset]))) & *((int *)&(filter->extractor.mask[0]))) !=\
			     *((int *)&(filter->value[0])) ) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			break;\
\
		EXTRACT_STRINGTYPE;\
\
		default: /* type is length: type <= XPF_DATA_MAXSIZE if (not one of above) is checked at filter-setup time */\
			type = xpfSizeFromType(type);\
			if (!(offset & 0x1) && (type & 0x1)) {\
				ret = XPF_DAGS_ERROR;	/* must be able to access a short around each offset */\
				goto end_compare;\
			}\
		avoid_buserror1:\
			if ( (offset < 0) || ((type + offset) >= (maxsize + 1)) ) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if ((ret = bmaskordercmp(&(buffer[offset]),\
						filter->value,\
						filter->extractor.mask,\
						type)) != 0) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			else break;\
		} /* switch */\
	}\
	else {	/* Must extract the pvalue from the buffer, and THEN try to hash it in the hash-table at the extractor: can't compare directly as above */\
		int		size;\
		HashEntry	**hashtable;\
		int		j;\
\
		if ((type = -type) == XPF_DATA_NONE) {\
			ret = XPF_DAGS_ERROR;	/* multivalued extractor can't matching everything: xpfInsertPatternList() */\
			goto end_compare;\
		}\
		offset = filter->extractor.selector.offset XPF_USER_GETCOUNTUNIT(filter);\
		NOPRINT(xpfPrintf("Compare: off %x type %x\n", offset, type);)\
		xpfBzero(pvalue, XPF_DATA_MAXSIZE);\
\
		switch(type)\
		{\
		case XPF_DATA_CHAR:\
			if ((offset < 0) || (offset >= maxsize)) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			if (offset&0x1)\
				pvalue[0] = *((short *)&buffer[offset-1]) & 0x00ff;\
			else {\
				if (maxsize-offset<2) {\
					ret = XPF_DAGS_ERROR;	/* must be able to access a short from around offset */\
					goto end_compare;\
				}\
				pvalue[0] = (*((short *)&buffer[offset]) & 0xff00) >> 8;\
			}\
			size = sizeof(char);\
			break;\
\
		case XPF_DATA_SHORT:\
			if ((offset < 0) || (offset >= maxsize + 1 - sizeof(short))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			*((short *)&(pvalue[0])) = ((*((short *)&(buffer[offset]))) & *((short *)&(filter->extractor.mask[0])));\
			size = sizeof(short);\
			break;\
\
		case XPF_DATA_INT:\
			if ((offset < 0) || (offset>= maxsize + 1 - sizeof(int))) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			*((int *)&(pvalue[0])) = ((*((int *)&(buffer[offset]))) & *((int *)&(filter->extractor.mask[0])));\
			size = sizeof(int);\
			break;\
\
		EXTRACT_STRINGTYPE_HASH\
\
		default: /* type is length: type <= XPF_DATA_MAXSIZE if (not one of above) is checked at filter-setup time */\
			size = type = xpfSizeFromType(type);\
			if (!(offset & 0x1) && (type & 0x1)) {\
				ret = XPF_DAGS_ERROR;	/* must be able to access a short around each offset */\
				goto end_compare;\
			}\
		avoid_buserror2:\
			if ( (offset < 0) || ((type + offset) >= (maxsize + 1)) ) {\
				ret = XPF_DAGS_ERROR;\
				goto end_compare;\
			}\
			memmaskcpy(pvalue, &buffer[offset], filter->extractor.mask, type);\
			if ((ret = bmaskordercmp(&(buffer[offset]),\
						filter->value,\
						filter->extractor.mask,\
						type)) != 0) {\
				ret = XPF_DAGS_AFTER;\
				goto end_compare;\
			}\
			else break;\
		} /* switch */\
\
		/* remove function calls: make inline code: hopefully, reduce single->multi valued timing jump */\
		index = xpfHashIndex(pvalue);\
\
		if ((hashtable = (HashEntry **)*(unsigned long *)(filter->value)) == NULL) {\
			ret = XPF_DAGS_AFTER;\
			goto end_compare;\
		}\
		pe = &hashtable[index];\
		while (*pe != NULL) {\
			/* LLP's suggestion: x-Kernel fundas about fixed size hash-keys\
			for (j=0; j<size; j++)\
				if ((*pe)->common.hash.value[j] != pvalue[j]) break;\
			if (j>=size) break;\
			*/\
			if ( *((unsigned long *)&(pvalue[0])) == *((unsigned long *)(*pe)->common.hash.value) ) break;\
			pe = &(*pe)->or;\
		}\
\
		if (*pe == NULL) {\
			ret = XPF_DAGS_AFTER;\
			goto end_compare;\
		}\
	}\
	ret = XPF_DAGS_SAME;\
end_compare:\
	;\
}
#endif	AVOIDBUSERROR
#endif	BYTEADDRESS

/*
 * Compares all the filters in the given pattern with elements in the
 * headers of protocols in the message.
 */
#if	DELAYEDMSGS
#define PROCESS_DELAYEDMSGS	if ((ret != XPF_DAGS_ERROR) && (p->flags & XPF_FLAG_DELAY)) return XPF_DAGS_DELAY;
#else	DELAYEDMSGS
#define PROCESS_DELAYEDMSGS
#endif	DELAYEDMSGS

#define xpfCompareWithPattern(state, message, length, pat, pe, ret)\
{\
	XPFFilter	*q;\
\
	q = (pat).filterlist;\
	xpfExtractHeader((state), (message), (length), (&(pat).header), (ret));\
	if (!ret) {\
		ret = XPF_DAGS_ERROR;\
		goto end_all;\
	}\
/* SAFETY\
	if ((state->xpfHeaderCurrent <= 0) ||\
	    (state->xpfHeaderData[state->xpfHeaderCurrent-1] == NULL) ||\
	    (state->xpfHeaderSizes[state->xpfHeaderCurrent-1] <= 0)) {\
		ret = XPF_DAGS_ERROR;\
		goto end_all;\
	}\
*/\
	while (q != NULL) {\
		pe = NULL;\
		xpfCompareWithFilter((state), (q), (pe), (ret));\
		if (ret != XPF_DAGS_SAME) {\
			PROCESS_DELAYEDMSGS\
			goto end_all;\
		}\
		/* MUST branch off: q->next MUST BE NULL now */\
		if ((pe != NULL) && (*pe != NULL)) {\
			ret = XPF_DAGS_SAME;\
			goto end_all;\
		}\
		q = q->next;\
	}\
\
	ret = XPF_DAGS_SAME;	/* Have traversed whole list w/o returning */\
end_all:\
	;\
}

/*
 * Returns a PATNode * at the end of the longest matching path of
 * patterns that match the given message. Used ONLY in xpf_interface.c/Lookup.
 *
 * Assumes and makes sure that its parameter proot always has nodetype
 * which is not a HASHENTRY: invariant. Its return value CAN be HASHENTRY.
 *
 * We need not have passed a PATNode ** in proot but we do for uniformity.
 * This function should be as efficient as possible: forget about others.
 * Therefore, I must remove recursion in this routine and make it a macro.
 */
#if	DELAYEDMSGS
#if	STOREMSGS
#define CHECK_DELAYEDMSGS\
	else if (ret == XPF_DAGS_DELAY) {\
		XPFMessageList	*m = (XPFMessageList *)xpfMalloc(sizeof(XPFMessageList));\
\
		m->next = (*proot)->messagelist;\
		(*proot)->messagelist = m;\
		m->message = message;\
		m->length = length;\
		state->xpfErrno = XPF_DAGS_DELAY;\
		xpfRestoreAll(state, message, length);\
		node = *proot;\
		goto bottom;\
	}
#else	STOREMSGS
#define CHECK_DELAYEDMSGS\
	else if (ret == XPF_DAGS_DELAY) {\
		state->xpfErrno = XPF_DAGS_DELAY;\
		xpfRestoreAll(state, message, length);\
		node = *proot;\
		goto bottom;\
	}
#endif	STOREMSGS
#else	DELAYEDMSGS
#define CHECK_DELAYEDMSGS
#endif	DELAYEDMSGS

#define xpfGetPATNodeFromMessage(state, message, length, ptrrootoftree, node)\
{\
	HashEntry	**pe;\
	PATNode		**proot;	/* ** since I have to set the matched nodes properly */\
	int		ret;\
\
	proot = ptrrootoftree;\
	if ((*proot == NULL) || ((*proot)->common.pattern.filterlist == NULL)) {\
		node = NULL;\
		goto bottom;\
	}\
top:\
	DOPRINT(xpfPrintf("In xpfGetPATNodeFromMessage: matchednodes = %x\n", state->xpfMatchedNum);)\
	DOPRINT(xpfPrintPATTree(*proot, 0);)\
	xpfCompareWithPattern((state), (message), (length), ((*proot)->common.pattern), (pe), (ret));\
	if (ret == XPF_DAGS_ERROR) {\
		node = NULL;\
		goto bottom;\
	}\
	CHECK_DELAYEDMSGS\
	else if (ret != XPF_DAGS_SAME) {\
		proot = &((*proot)->or);\
		if ((*proot == NULL) || ((*proot)->common.pattern.filterlist == NULL)) {\
			node = (state->xpfMatchedNum > 0) ? *state->xpfMatchedNodes[state->xpfMatchedNum - 1] : NULL;\
			goto bottom;\
		}\
		goto top;\
	}\
	else {	/* They are the same */\
		state->xpfMatchedNodes[state->xpfMatchedNum ++] = proot;\
		if ((pe == NULL) || (*pe == NULL)) {\
			proot = &((*proot)->and);\
			if ((*proot == NULL) || ((*proot)->common.pattern.filterlist == NULL)) {\
				node = (state->xpfMatchedNum > 0) ? *state->xpfMatchedNodes[state->xpfMatchedNum - 1] : NULL;\
				goto bottom;\
			}\
			goto top;\
		}\
		else {\
			state->xpfMatchedNodes[state->xpfMatchedNum ++] = pe;	\
			proot = &((*pe)->and);\
			if ((*proot == NULL) || ((*proot)->common.pattern.filterlist == NULL)) {\
				node = (state->xpfMatchedNum > 0) ? *state->xpfMatchedNodes[state->xpfMatchedNum - 1] : NULL;\
				goto bottom;\
			}\
			goto top;\
		}\
	}\
bottom:\
	;\
}

#endif	_XPF_INTERNAL_H_
